/*
 * scdis:  MSFS Scenery Decompiler for scasm.
 *
 * Copyright (C) 1996-2000 Takuya Murakami. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: sec0ana.c,v 1.3 2000/01/04 06:41:18 tmurakam Exp $
 */
/*
 * Section 0 (VOR, ILS) analyze
 */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "type.h"
#include "util.h"
#include "conv.h"
#include "sec0ana.h"


#define MAXLRANGE	500

/* global variables */
static lat_int_t	minlats[MAXLRANGE];
static lat_int_t	maxlats[MAXLRANGE];
static int nrange;

/* prototypes */
void AnalyzeFreqTab(u_int32, u_int32, double);
void AnalyzeVORILS(u_int32, double);
void anaVOR(double);
void anaILS(double);
void ClearLatRange(void);
int OpenSec0File(lat_int_t, lat_int_t);

extern void
AnalyzeSec0(u_int32 sec_pointer, double first_freq, double max_freq)
{
	u_int32 pos = sec_pointer;
	u_int8 c;
	u_int32	offset;
	int	i = 0;

	SET_PASS2;	/* no PASS 1 */

	nrange = 1;

	MovePointer(pos);

	while (1) {
		c = get1();
		if (c == 0) break;

		switch (c) {
		case 0x01:
			offset = get4();
			pos = GetPointer();

			if (offset) {
				AnalyzeFreqTab(sec_pointer + offset, sec_pointer, 
					first_freq + 0.05 * i);
			}

			i++;

			MovePointer(pos);
			break;

		default:
			fprintf(stderr, "unknown record\n");
			exit(1);
		}

		if (first_freq + 0.05 * i > max_freq + 0.01) break; /* ### ad hoc */
	}
}

void AnalyzeFreqTab(u_int32 freqtab, u_int32 basepos, double freq)
{
	u_int32 pos = freqtab;
	u_int8	c;
	lat_int_t	minlat, maxlat;
	u_int32	offset;

	MovePointer(freqtab);

	while (1) {
		c = get1();
		if (c == 0) break;

		switch (c) {
		case 0x15:
			minlat = (u_int32)get2() << 8;
			maxlat = (u_int32)get2() << 8;
			offset = get4();

			if (OpenSec0File(minlat, maxlat)) {
				/* Write Header and LatRange */
				WriteHeader(0);

				fprintf(outfp, "LatRange( %s ", latstr(lat_int2lat(minlat), NULL));
				fprintf(outfp, "%s )\n", latstr(lat_int2lat(maxlat), NULL));
			}

			pos = GetPointer();
			AnalyzeVORILS(basepos + offset, freq);
			MovePointer(pos);

			CloseOutFile();
			break;

		default:
			fprintf(stderr, "unknown record\n");
			exit(1);
		}
	}
}

void
ClearLatRange(void)
{
	nrange = 0;
}

int
OpenSec0File(lat_int_t minlat, lat_int_t maxlat)
{
	int i;
	char fname[30];

	for (i=0; i<nrange; i++) {
		if (minlat == minlats[i] && maxlat == maxlats[i]) {
			break;
		}
	}
	if (i >= nrange) {
		/* open new file */
		sprintf(fname, "%s0-%d", SECFILENAME_PRE, nrange);
		OpenOutFile(fname, TRUE);
		minlats[nrange] = minlat;
		maxlats[nrange] = maxlat;
		nrange++;
		return 1;
	} else {
		sprintf(fname, "%s0-%d.scm", SECFILENAME_PRE, i);
		AppendOutFile(fname);
		return 0;
	}
}



void
AnalyzeVORILS(u_int32 pos, double freq)
{
	u_int8	c;

	MovePointer(pos);

	while (1) {
		c = get1();
		if (c == 0) break;

		switch (c) {
		case 0x04:
			anaVOR(freq);
			break;

		case 0x05:
			anaILS(freq);
			break;

		case 0x14:
			/* unknown record */
			fprintf(outfp, ";;; unknown record (0x14)\n");
			fprintf(outfp, ";;; ");
			Dump(17);
			break;

		default:
			fprintf(outfp, ";;; unknown record (0x%x)\n", c);
			break;
		}
	}
}


/*-- 04h ------------- VOR (44 bytes)


1     1   Range in 2048 m units
2     2   Magnetic deviation (in Deg16)
4     1   Flags; bit     0 DME no/yes
                         1 ??
          Source:        3 unused?
          JOHN TRINDLE   4 unused?
                         5 unused?
                         6 glideslope: 0=no (VOR) 1=yes (ILS)
                         7 sensitivity: 0=ADF/VOR 1=ILS
5     3   Latitude  in INu
8     3   Longitude in IEu
11    2   Altitude  in m
13    2   approach course (used only for ILSes, always 0 for VORs)
15    5   VOR ID; string padded with spaces
20    24  VOR Name; string padded with spaces*/

void
anaVOR(double freq)
{
	u_int8		range, flag;
	double		mag;
	lat_int_t	lat;
	lon_int_t	lon;
	u_int16		alt;
	double		course;
	char		id[6];
	char		name[25];

	range = get1();
	mag = deg162deg(get2());
	flag = get1();

	lat  = get3();
	lon  = get3();

	alt = get2();
	course = deg162deg(get2());

	getlong(id, 5);
	id[5] = 0;
	NormalString(id);

	getlong(name, 24);
	name[24] = 0;
	NormalString(name);

	if (IS_PASS2) {
		fprintf(outfp, "\tVOR( %.2f %d %02X %.2f %s %s\n", 
			freq, range, flag, mag, id, name);
		fprintf(outfp, "\t\t%s ", latstr(lat_int2lat(lat), NULL));
		fprintf(outfp, "%s ", lonstr(lon_int2lon(lon), NULL));
		fprintf(outfp, "%d %.2f )\n", alt, course);
	}
}


/*-- 05h ------------- ILS (54 bytes)

1     1   Range in 2048 m units
2     2   Magnetic deviation (in Deg16)
4     1   Flags (see rec. 04h above)
5     3   Latitude  in INu
8     3   Longitude in IEu
11    2   Altitude  in m
13    2   Approach course (in Deg16)
15    5   VOR ID; string padded with spaces
20    24  VOR Name; string padded with spaces
44    3   Glideslope latitude  in INu
47    3   Glideslope longitude in IEu
50    2   Glideslope altitude  in m
52    2   Glideslope angle?
*/

void
anaILS(double freq)
{
	u_int8		range, flag;
	double		mag;
	lat_int_t	lat, glat;
	lon_int_t	lon, glon;
	u_int16		alt, galt;
	double		course, gcourse;
	char		id[6];
	char		name[25];

	range = get1();
	mag = deg162deg(get2());
	flag = get1();

	lat  = get3();
	lon  = get3();

	alt = get2();
	course = deg162deg(get2());

	getlong(id, 5);
	id[5] = 0;
	NormalString(id);

	getlong(name, 24);
	name[24] = 0;
	NormalString(name);

	glat  = get3();
	glon  = get3();

	galt = get2();
	/*gcourse = deg162deg(get2());*/
	gcourse = (double)get2() / 65536.0 / 2.0 * 180 / 3.14159265359;	/* ad hoc ... ### */

	if (IS_PASS2) {
		fprintf(outfp, "\tILS( %.2f %d %02X %.2f %s %s\n", 
			freq, range, flag, mag, id, name);
		
		fprintf(outfp, "\t\t%s ", latstr(lat_int2lat(lat), NULL));
		fprintf(outfp, "%s ", lonstr(lon_int2lon(lon), NULL));
		fprintf(outfp, "%d %.2f\n", alt, course);

		fprintf(outfp, "\t\t%s ", latstr(lat_int2lat(glat), NULL));
		fprintf(outfp, "%s ", lonstr(lon_int2lon(glon), NULL));
		fprintf(outfp, "%d %.2f )\n", galt, gcourse);
	}
}


